@@ -6,16 +6,22 @@ import trajectory_planning_helpers as tph
 import pandas as pd
 
 
+"""
+    Python version: 3.7
+    Created by: Tim Stahl (based on work by George Willis & Leonhard Hermansdorfer)
+    Created on: 18.02.2020
+
+    This package takes track bounds (coordinate arrays) and generates a reference line, including matching
+    normal-vectors and track-widths.
+"""
+
+
 def generate_refline(bound_l: np.ndarray,
                      bound_r: np.ndarray,
                      resolution: float = 1.0):
     """
-    Python version: 3.5
-    Created by: Tim Stahl (based on work by George Willis & Leonhard Hermansdorfer)
-    Created on: 18.02.2020
-
-    Documentation:  This function takes track bounds and generates a reference line, including corresponding
-                    normal-vectors and track-widths.
+    This function takes track bounds and generates a reference line, including corresponding normal-vectors and
+    track-widths.
 
     :param bound_l:         bound coords left given as numpy array (columns holding x and y coords)
     :param bound_r:         bound coords right given as numpy array (columns holding x and y coords)
@@ -39,7 +45,7 @@ def generate_refline(bound_l: np.ndarray,
 
     # get current distances and interpolate to desired resolution (distance of points)
     distances, distances_sum = get_distance(input_coords=center_line)
-    iter_track = interpol_new_points(input_coords=center_line,
+    iter_track = interpol_new_points(input_coord=center_line,
                                      distances=distances,
                                      track_res=resolution)
 
@@ -73,98 +79,108 @@ def generate_refline(bound_l: np.ndarray,
 
     # -- iteratively smooth until track bound is left ------------------------------------------------------------------
     bool_check = inpolygon(track_polygon=track_polygon,
-                           input_rl=iter_track)
+                           input_coord=iter_track)
     n = 0
     temp_iter_track = np.copy(iter_track)
     while bool_check and not n > 500:
-        temp_iter_track = movmean_smoothing(input_rl=temp_iter_track,
+        temp_iter_track = movmean_smoothing(input_coord=temp_iter_track,
                                             window_size=9)
         n += 1
         if n % 5 == 0:
             bool_check = inpolygon(track_polygon=track_polygon,
-                                   input_rl=temp_iter_track)
+                                   input_coord=temp_iter_track)
 
             if bool_check:
                 iter_track = np.copy(temp_iter_track)
 
     # - split into sections (based on heading changes) -----------------------------------------------------------------
-    heading_angle = get_heading_angle(input_rl=iter_track)
+    heading_angle = get_heading_angle(input_coord=iter_track)
     section_row_indices = find_sections(heading_angle=heading_angle)[0]
 
     # -- smooth all sections individually until no overlapping normal-vectors exist ------------------------------------
-    output_rl = smooth_sections(input_rl=iter_track,
+    output_rl = smooth_sections(input_coord=iter_track,
                                 section_row_indices=section_row_indices,
                                 track_polygon=track_polygon)
-    heading_angle = get_heading_angle(input_rl=output_rl)
-    output_rl = final_smooth(input_rl=output_rl,
+    heading_angle = get_heading_angle(input_coord=output_rl)
+    output_rl = final_smooth(input_coord=output_rl,
                              heading_angle=heading_angle)
 
     # -- generate outputs ----------------------------------------------------------------------------------------------
     # calculate normal vectors
-    v_normal = create_normal_vectors(input_rl=output_rl)
+    v_normal = create_normal_vectors(input_coord=output_rl)
 
     # get intersection points with bounds (in order to determine track-width)
     intersects_left, intersects_right = get_intersect_points(track_polygon=track_polygon,
-                                                             input_rl=output_rl,
+                                                             input_coord=output_rl,
                                                              v_normal=v_normal)
 
     # calculate track-width based on intersection points
-    tw_left, tw_right = get_track_width(input_rl=output_rl,
+    tw_left, tw_right = get_track_width(input_coord=output_rl,
                                         intersects_left=intersects_left,
                                         intersects_right=intersects_right)
 
     return output_rl, v_normal, tw_left, tw_right
 
 
-def get_distance(input_coords: np.ndarray):
+def get_distance(input_coords: np.ndarray) -> tuple:
     """
-    Get the distances between each point in 'input_coords' in meters
+    Calculates distances between earch point in 'input_coords' and returns a array of these distances as well as an
+    cumulative row, starting at 0.0.
+
+    :param input_coords:        input coordinates (columns holding x and y coordinates)
+    :returns (distances,        distances between the 'input_coords' in m
+              distances_sum)    cumulative sequence of the 'distances' starting at 0.0
     """
 
     # Get euclidean distances
     distances = np.sqrt(np.sum(np.power(np.diff(input_coords, axis=0), 2), axis=1))
 
     # Cumulative sum of the distance vector
-    distances_sum = np.append([[0]], np.cumsum(distances))
+    distances_sum = np.append([[0.0]], np.cumsum(distances))
 
     return distances, distances_sum
 
 
-def interpol_new_points(input_coords: np.ndarray,
+def interpol_new_points(input_coord: np.ndarray,
                         distances: np.ndarray,
-                        track_res: float):
+                        track_res: float) -> np.ndarray:
     """
-    Interpolate new points to the given resolution in the main script
+    Interpolate new points to the given track resolution.
+
+    :param input_coord:     input coordinates (columns holding x and y coordinates)
+    :param distances:       distances between the coordinates 'input_coords'
+    :param track_res:       desired track resolution in m
+    :returns coords_out:    returned (interpolated) coordinates (columns holding x and y coordinates)
     """
 
     # -- step 1 - interpolate points with very tight spacing -----------------------------------------------------------
-    new_points = np.zeros((1, 2))
+    coords_out = np.zeros((1, 2))
     dist_limit = 0.005  # maximum distances between two coordinates
-    v_length = len(input_coords[:, 0])
+    v_length = len(input_coord[:, 0])
     for row in range(v_length - 1):
-        p1 = input_coords[row, :]  # Get both points from the input
-        p2 = input_coords[row + 1, :]
+        p1 = input_coord[row, :]  # Get both points from the input
+        p2 = input_coord[row + 1, :]
         if distances[row] < dist_limit:
             # if the dist is smaller than the limit --> no interpol needed
             interpl_points = np.array([p1])
-            new_points = np.append(new_points, interpl_points, axis=0)
+            coords_out = np.append(coords_out, interpl_points, axis=0)
         else:
             # if distance larger insert points to reduce the distance
             num_new_points = math.ceil(distances[row] / dist_limit)
             interpl_points = np.linspace(p1, p2, num=num_new_points, endpoint=False)
-            new_points = np.append(new_points, interpl_points, axis=0)
+            coords_out = np.append(coords_out, interpl_points, axis=0)
 
     # add last end-point and remove first point (initialized to zero)
-    new_points = np.append(new_points, np.array([input_coords[-1, :]]), axis=0)
-    new_points = np.delete(new_points, 0, axis=0)
+    coords_out = np.append(coords_out, np.array([input_coord[-1, :]]), axis=0)
+    coords_out = np.delete(coords_out, 0, axis=0)
 
     # -- step 2 - remove points in order to obtain desired step width --------------------------------------------------
-    indices = np.zeros((len(new_points), 1))
+    indices = np.zeros((len(coords_out), 1))
     row1 = 0
     row2 = 1
-    while row1 + row2 < len(new_points):
-        delta_x = new_points[row1, 0] - new_points[row1 + row2, 0]
-        delta_y = new_points[row1, 1] - new_points[row1 + row2, 1]
+    while row1 + row2 < len(coords_out):
+        delta_x = coords_out[row1, 0] - coords_out[row1 + row2, 0]
+        delta_y = coords_out[row1, 1] - coords_out[row1 + row2, 1]
         distance = np.sqrt(delta_x ** 2 + delta_y ** 2)
         if distance < track_res:
             indices[row1 + row2] = 1
@@ -174,23 +190,26 @@ def interpol_new_points(input_coords: np.ndarray,
             row2 = 1
     indices[-1] = 0
     idx = np.where(indices[:, 0] == 1)
-    new_points = np.delete(new_points, idx, axis=0)
+    coords_out = np.delete(coords_out, idx, axis=0)
 
-    return new_points
+    return coords_out
 
 
-def get_heading_angle(input_rl: np.ndarray):
+def get_heading_angle(input_coord: np.ndarray):
     """
-    Heading vector (i.e. the angle of a single-track wheel following the trajectory)
-    Value at certain index represents the heading angle at that given point (i.e. angle betw. 2 vectors in that point)
+    Calculates heading vector (i.e. the angle of a single-track wheel following the trajectory).
+    Value at certain index represents the heading angle at that given point (i.e. angle betw. 2 vectors in that point).
+
+    :param input_coord:     input coordinates (columns holding x and y coordinates)
+    :returns heading_angle: returned heading angle along the path
     """
 
-    heading_angle = np.zeros((len(input_rl[:, 0]), 1))
-    for row in range(1, len(input_rl) - 1):
+    heading_angle = np.zeros((len(input_coord[:, 0]), 1))
+    for row in range(1, len(input_coord) - 1):
         heading_angle[row, 0] = np.degrees(
-            tph.normalize_psi.normalize_psi(tph.angle3pt.angle3pt(a=input_rl[row - 1, :],
-                                                                  b=input_rl[row, :],
-                                                                  c=input_rl[row + 1, :]) - math.pi)
+            tph.normalize_psi.normalize_psi(tph.angle3pt.angle3pt(a=input_coord[row - 1, :],
+                                                                  b=input_coord[row, :],
+                                                                  c=input_coord[row + 1, :]) - math.pi)
         )
 
     return heading_angle
@@ -201,14 +220,18 @@ def get_angle_diff(heading_angle: np.ndarray):
 
 
 def inpolygon(track_polygon: Polygon,
-              input_rl: np.ndarray):
+              input_coord: np.ndarray) -> bool:
     """
-    Check if all coordinates are within the track polygon
+    Checks if all provided coordinates lie within the track bounds (polygon).
+
+    :param track_polygon:   track bounds represented by a closed polygon
+    :param input_coord:     input coordinates (columns holding x and y coordinates)
+    :returns in_polygon:    'True' if all points lie within polygon, 'False' if any coordinate is outside
     """
 
-    bool_check_list = np.zeros((len(input_rl[:, 0]), 1), dtype=bool)
-    for row in range(len(input_rl)):
-        p = Point(input_rl[row, :])
+    bool_check_list = np.zeros((len(input_coord[:, 0]), 1), dtype=bool)
+    for row in range(len(input_coord)):
+        p = Point(input_coord[row, :])
 
         # check if point is within polygon
         bool_check_list[row, :] = p.within(track_polygon)
@@ -216,17 +239,22 @@ def inpolygon(track_polygon: Polygon,
     return np.all(bool_check_list)
 
 
-def movmean_smoothing(input_rl: np.ndarray,
-                      window_size: int):
+def movmean_smoothing(input_coord: np.ndarray,
+                      window_size: int) -> np.ndarray:
     """
-    Smoothing function based on the moving average function. The first and last coordinates always stay the same.
+    Smooth coordinate line based on moving average function. The first and last coordinates always stay the same.
+
+    :param input_coord:     input coordinates (columns holding x and y coordinates)
+    :param window_size:     number of points to be included for the moving average calculation
+    :returns output_coord:  smoothed output coordinates (columns holding x and y coordinates)
     """
-    input_save = input_rl
+
+    input_save = input_coord
     # Get the distances of the input
-    _, distances_sum = get_distance(input_rl)
+    _, distances_sum = get_distance(input_coord)
     # Convert the numpy array into two DataFrames
-    x = pd.DataFrame(input_rl[:, 0])
-    y = pd.DataFrame(input_rl[:, 1])
+    x = pd.DataFrame(input_coord[:, 0])
+    y = pd.DataFrame(input_coord[:, 1])
 
     # Perform a moving average on the x and y column independent from each other with the given window size
     sx = x.rolling(window=window_size)
@@ -239,16 +267,16 @@ def movmean_smoothing(input_rl: np.ndarray,
     sy = sy.to_numpy()
 
     # Create the the original 2D Array
-    output_rl = np.hstack((sx, sy))
+    output_coord = np.hstack((sx, sy))
     # Replace the NaNs and last array with a smaller smoothing window to keep
     if window_size == 3:
-        output_rl = np.delete(output_rl, [0, 1], axis=0)
+        output_coord = np.delete(output_coord, [0, 1], axis=0)
         first_point = input_save[0, :]
         last_point = input_save[-1, :]
-        output_rl = np.vstack((first_point, output_rl))
-        output_rl = np.vstack((output_rl, last_point))
+        output_coord = np.vstack((first_point, output_coord))
+        output_coord = np.vstack((output_coord, last_point))
     if window_size == 9:
-        output_rl = np.delete(output_rl, [0, 1, 2, 3, 4, 5, 6, 7], axis=0)
+        output_coord = np.delete(output_coord, [0, 1, 2, 3, 4, 5, 6, 7], axis=0)
         append_top = input_save[0, :]
         append_bottom = input_save[-1, :]
         for i in range(3, 8, 2):
@@ -257,15 +285,20 @@ def movmean_smoothing(input_rl: np.ndarray,
         for i in range(-3, -8, -2):
             p = np.mean(input_save[i:, :], axis=0)
             append_bottom = np.vstack((p, append_bottom))
-        output_rl = np.vstack((append_top, output_rl))
-        output_rl = np.vstack((output_rl, append_bottom))
+        output_coord = np.vstack((append_top, output_coord))
+        output_coord = np.vstack((output_coord, append_bottom))
 
-    return output_rl
+    return output_coord
 
 
-def find_sections(heading_angle: np.ndarray):
+def find_sections(heading_angle: np.ndarray) -> tuple:
     """
-    After initial smoothing, find all sections where the angle change its sign (i.e. left turn goes to right turn)
+    Find all sections where the angle changes its sign (i.e. left turn goes to right turn)
+
+    :param heading_angle:           array of heading course
+    :returns (section_row_indices,  indices indicating a sign change within the 'heading_angle' array
+              section_angles,       angles at the positions of 'section_row_indices'
+              delete_row_indices):  indices where no sign change within the 'heading_angle' array occurs
     """
 
     # if the sign changes change the array element to 1
@@ -286,34 +319,46 @@ def find_sections(heading_angle: np.ndarray):
     return section_row_indices, section_angles, delete_row_indices
 
 
-def final_smooth(input_rl, heading_angle):
+def final_smooth(input_coord: np.ndarray,
+                 heading_angle: np.ndarray) -> np.ndarray:
+    """
+    Smooth path coordinates until maximal angle alternation does not decline any further.
+
+    :param input_coord:       input coordinates (columns holding x and y coordinates)
+    :param heading_angle:     array of heading course
+    :returns output_coord:    smoothed output coordinates (columns holding x and y coordinates)
+    """
+
     section_angles_diff_in = max(abs(get_angle_diff(heading_angle)))
-    section_angles_diff_out = 0
+    section_angles_diff_out = 0.0
     while section_angles_diff_in > section_angles_diff_out:
-        heading_angle = get_heading_angle(input_rl)
+        heading_angle = get_heading_angle(input_coord)
         section_angles_diff_in = max(abs(get_angle_diff(heading_angle)))
-        input_rl = movmean_smoothing(input_rl, 9)
-        heading_angle_out = get_heading_angle(input_rl)
+        input_coord = movmean_smoothing(input_coord, 9)
+        heading_angle_out = get_heading_angle(input_coord)
         section_angles_diff_out = max(abs(get_angle_diff(heading_angle_out)))
         if section_angles_diff_in - section_angles_diff_out < 0.01:
             break
 
-    output_rl = input_rl
-    return output_rl
+    return input_coord
 
 
-def create_normal_vectors(input_rl: np.ndarray):
+def create_normal_vectors(input_coord: np.ndarray) -> np.ndarray:
     """
-    Get the normal vectors, which extend 90 degrees from each coordinate pair
+    Get the normal vectors, which extend orthogonal to each coordinate pair.
+
+    :param input_coord:     input coordinates (columns holding x and y coordinates)
+    :returns v_normal:      normal vectors (columns holding x and y components)
     """
+
     # extrapolate the direction last vector
     # Get the x and y direction of the last vector
-    d_x = input_rl[-1, 0] - input_rl[-2, 0]
-    d_y = input_rl[-1, 1] - input_rl[-2, 1]
+    d_x = input_coord[-1, 0] - input_coord[-2, 0]
+    d_y = input_coord[-1, 1] - input_coord[-2, 1]
     # The new coordinate is the last coordinate plus d_x and d_y
-    last_vector = np.array((input_rl[-1, 0] + d_x, input_rl[-1, 1] + d_y))
+    last_vector = np.array((input_coord[-1, 0] + d_x, input_coord[-1, 1] + d_y))
     # Add the new coordinate to the array end
-    output_rl = np.vstack((input_rl, last_vector))
+    output_rl = np.vstack((input_coord, last_vector))
 
     # create the orthogonal vectors left and right for each vector
     v_length = len(output_rl[:, 0])
@@ -336,17 +381,24 @@ def create_normal_vectors(input_rl: np.ndarray):
 
 
 def get_intersect_points(track_polygon: Polygon,
-                         input_rl: np.ndarray,
-                         v_normal: np.ndarray):
+                         input_coord: np.ndarray,
+                         v_normal: np.ndarray) -> tuple:
     """
-    Find the coordinates where the extended normal vectors cross ith the track boundaries
+    Find the coordinates where the extended normal vectors cross the track boundaries.
+
+    :param track_polygon:       track bounds represented by a closed polygon
+    :param input_coord:         input coordinates (columns holding x and y coordinates)
+    :param v_normal:            normal vectors (columns holding x and y components)
+    :returns (intersects_left,  intersection coordinates to the left of input_coord (columns holding x and y coords)
+              intersects_right) intersection coordinates to the right of input_coord (columns holding x and y coords)
     """
+
     # Extend normal vector to 50.0 meters (definitely hit borders) and add it to race-line
-    v_normal_left = input_rl - v_normal * 50.0
-    v_normal_right = input_rl + v_normal * 50.0
+    v_normal_left = input_coord - v_normal * 50.0
+    v_normal_right = input_coord + v_normal * 50.0
 
     # create the output array with zeros
-    v_length = len(input_rl[:, 0])
+    v_length = len(input_coord[:, 0])
     intersects_left = np.zeros((v_length, 2))
     intersects_right = np.zeros((v_length, 2))
 
@@ -354,7 +406,7 @@ def get_intersect_points(track_polygon: Polygon,
     for row in range(v_length):
         # Left Vectors:
         # Create a Shapely Line with the end point of the normal vector and the reference line coordinates
-        line_left = LineString([input_rl[row, :], v_normal_left[row, :]])
+        line_left = LineString([input_coord[row, :], v_normal_left[row, :]])
         # Get the coordinate where the new line crosses the track boundaries
         intersect_left_iter = track_polygon.intersection(line_left)
         # Get the coordinates out the Multipoint data type
@@ -371,7 +423,7 @@ def get_intersect_points(track_polygon: Polygon,
             intersects_left[row, :] = np.array(left_coords_values)
 
         # Same procedure, but for the right side
-        line_right = LineString([input_rl[row, :], v_normal_right[row, :]])
+        line_right = LineString([input_coord[row, :], v_normal_right[row, :]])
         intersect_right_iter = track_polygon.intersection(line_right)
         right_coords_dict = intersect_right_iter.__geo_interface__
         right_coords_values = right_coords_dict.get('coordinates')
@@ -386,26 +438,37 @@ def get_intersect_points(track_polygon: Polygon,
     return intersects_left, intersects_right
 
 
-def get_track_width(input_rl: np.ndarray,
+def get_track_width(input_coord: np.ndarray,
                     intersects_left: np.ndarray,
-                    intersects_right: np.ndarray):
+                    intersects_right: np.ndarray) -> tuple:
     """
-    Calculate the distances for input reference line and the intersects with the track bounds left and right
+    Calculate the distances from input coordinates and the intersects with the track bounds left and right (track-width
+    at corresponding point).
+
+    :param input_coord:          input coordinates (columns holding x and y coordinates)
+    :param intersects_left       intersection coordinates to the left of input_coord (columns holding x and y coords)
+    :param intersects_right      intersection coordinates to the right of input_coord (columns holding x and y coords)
+    :returns (track_width_left,  distances between input coordinates and left intersection coordinates
+              track_width_right) distances between input coordinates and right intersection coordinates
     """
 
-    track_width_left = np.sqrt(np.sum(np.power(input_rl - intersects_left, 2), axis=1))[np.newaxis].T
-    track_width_right = np.sqrt(np.sum(np.power(input_rl - intersects_right, 2), axis=1))[np.newaxis].T
+    track_width_left = np.sqrt(np.sum(np.power(input_coord - intersects_left, 2), axis=1))[np.newaxis].T
+    track_width_right = np.sqrt(np.sum(np.power(input_coord - intersects_right, 2), axis=1))[np.newaxis].T
 
     return track_width_left, track_width_right
 
 
 def normal_vectors_intersections(intersects_left: np.ndarray,
-                                 intersects_right: np.ndarray):
+                                 intersects_right: np.ndarray) -> tuple:
     """
     Check if neighboring normal-vectors cross each other within the track-area.
 
-    Output: Row indices of valid and invalid lines
+    :param intersects_left          intersection coordinates to the left of ref. line (columns holding x and y coords)
+    :param intersects_right         intersection coordinates to the right of ref. line (columns holding x and y coords)
+    :returns (rows_with_intersect,  (row) indices of valid normal-vectors
+              delete_row_indices)   (row) indices of invalid normal-vectors
     """
+
     # Array with zeros to improve speed
     v_length = len(intersects_left[:, 0])
     bool_intersect = np.zeros((v_length, 1))
@@ -433,12 +496,21 @@ def normal_vectors_intersections(intersects_left: np.ndarray,
     return rows_with_intersect, delete_row_indices
 
 
-def smooth_sections(input_rl: np.ndarray,
+def smooth_sections(input_coord: np.ndarray,
                     section_row_indices: list,
-                    track_polygon: Polygon):
+                    track_polygon: Polygon) -> np.ndarray:
+    """
+    Smooth coordinate line within sections.
+
+    :param input_coord:         input coordinates (columns holding x and y coordinates)
+    :param section_row_indices  indices indicating a sign change within the 'heading_angle' array
+    :param track_polygon:       track bounds represented by a closed polygon
+    :returns coords_out:        returned (smoothed) coordinates (columns holding x and y coordinates)
+    """
+
     num_sections = len(section_row_indices)  # Get the number of sections within the track
-    length_input = len(input_rl)  # Get the number of input reference line coordinates
-    output_rl = np.zeros((length_input, 2))  # Make all output arrays the same size with zeros
+    length_input = len(input_coord)  # Get the number of input reference line coordinates
+    output_coord = np.zeros((length_input, 2))  # Make all output arrays the same size with zeros
     section_row_indices = np.hstack((section_row_indices, [length_input]))
 
     # print('sections started')
@@ -456,7 +528,7 @@ def smooth_sections(input_rl: np.ndarray,
             sec_start = section_row_indices[row]
             sec_end = section_row_indices[row + 1]
 
-        section = input_rl[sec_start:sec_end, :]
+        section = input_coord[sec_start:sec_end, :]
 
         # Are there any intersects in the section
         v_normal = create_normal_vectors(section)
@@ -468,7 +540,7 @@ def smooth_sections(input_rl: np.ndarray,
 
         # If no intersects exist from the beginning, continue to the next section
         if rows_with_intersect.dtype == bool and angle_bool is True:
-            output_rl[sec_start:sec_end, :] = section
+            output_coord[sec_start:sec_end, :] = section
             # print('section ' + str(row + 1) + ' of ' + str(num_sections) + ' finished')
             continue
 
@@ -478,15 +550,15 @@ def smooth_sections(input_rl: np.ndarray,
             v_normal = create_normal_vectors(section)
             intersects_left, intersects_right = get_intersect_points(track_polygon, section, v_normal)
             rows_with_intersect, _ = normal_vectors_intersections(intersects_left, intersects_right)
-        output_rl[sec_start:sec_end, :] = section
+        output_coord[sec_start:sec_end, :] = section
 
-    return output_rl
+    return output_coord
 
 
 # -- main --------------------------------------------------------------------------------------------------------------
 if __name__ == "__main__":
     scenario_path = (os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
-                     + "/sample_scenarios/n_vehicle/modena_overtake_tight.scn")
+                     + "/sample_files/scenario_n_vehicle/modena_overtake_tight.scn")
 
     import scenario_testing_tools
     b_l, b_r = scenario_testing_tools.get_scene_track.get_scene_track(file_path=scenario_path)
